@@ -19,7 +19,7 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  *
- * $Id: gtkimcontextime.c,v 1.8 2003/05/07 12:42:50 makeinu Exp $
+ * $Id: gtkimcontextime.c,v 1.11 2003/05/09 14:56:35 makeinu Exp $
  */
 
 #include "gtkimcontextime.h"
@@ -56,27 +56,8 @@ struct _GtkIMContextIMEPrivate
 };
 
 
-/* #define IMPLEMENT_GET_PREEDIT_STRING 1 */
 #define BUFSIZE 4096
 
-#define GET_UTF8_PREEDIT_STRING(himc, idx, str)              \
-{                                                            \
-    gchar buf[BUFSIZE];                                      \
-    glong len;                                               \
-    gsize bytes_read, bytes_written;                         \
-    GError *error = NULL;                                    \
-                                                             \
-    len = ImmGetCompositionString (himc, idx,                \
-                                   buf, G_N_ELEMENTS (buf)); \
-    str = g_locale_to_utf8 (buf, len,                        \
-                            &bytes_read, &bytes_written,     \
-                            &error);                         \
-    if (error) {                                             \
-        g_warning ("%s", error->message);                    \
-        g_error_free (error);                                \
-    }                                                        \
-}
-
 #define FREE_PREEDIT_BUFFER(ctx)   \
 {                                  \
     g_free((ctx)->priv->comp_str); \
@@ -98,12 +79,10 @@ static void     gtk_im_context_ime_set_client_window   (GtkIMContext   *context,
 static gboolean gtk_im_context_ime_filter_keypress     (GtkIMContext   *context,
                                                         GdkEventKey    *event);
 static void     gtk_im_context_ime_reset               (GtkIMContext   *context);
-#ifdef IMPLEMENT_GET_PREEDIT_STRING
 static void     gtk_im_context_ime_get_preedit_string  (GtkIMContext   *context,
                                                         gchar         **str,
                                                         PangoAttrList **attrs,
                                                         gint           *cursor_pos);
-#endif /* IMPLEMENT_GET_PREEDIT_STRING */
 static void     gtk_im_context_ime_focus_in            (GtkIMContext   *context);
 static void     gtk_im_context_ime_focus_out           (GtkIMContext   *context);
 static void     gtk_im_context_ime_set_cursor_location (GtkIMContext   *context,
@@ -154,9 +133,7 @@ gtk_im_context_ime_class_init (GtkIMContextIMEClass *class)
     im_context_class->set_client_window   = gtk_im_context_ime_set_client_window;
     im_context_class->filter_keypress     = gtk_im_context_ime_filter_keypress;
     im_context_class->reset               = gtk_im_context_ime_reset;
-#ifdef IMPLEMENT_GET_PREEDIT_STRING
     im_context_class->get_preedit_string  = gtk_im_context_ime_get_preedit_string;
-#endif /* IMPLEMENT_GET_PREEDIT_STRING */
     im_context_class->focus_in            = gtk_im_context_ime_focus_in;
     im_context_class->focus_out           = gtk_im_context_ime_focus_out;
     im_context_class->set_cursor_location = gtk_im_context_ime_set_cursor_location;
@@ -170,8 +147,16 @@ static void
 gtk_im_context_ime_init (GtkIMContextIME *context_ime)
 {
     GdkRectangle *area;
+    const gchar *mode;
+
+    mode = g_getenv ("GTK_IM_CONTEXT_IME_PREEDIT_MODE");
 
     context_ime->client_window = NULL;
+    if (mode && !strcasecmp(mode, "GTK")) {
+        context_ime->preedit_mode  = GTK_IM_CONTEXT_IME_PREEDIT_GTK;
+    } else {
+        context_ime->preedit_mode  = GTK_IM_CONTEXT_IME_PREEDIT_IME;
+    }
     context_ime->preediting    = FALSE;
     context_ime->opened        = FALSE;
 
@@ -208,6 +193,56 @@ gtk_im_context_ime_new (void)
 }
 
 
+static gchar *
+get_preedit_string (HIMC himc, DWORD idx, gint *pos)
+{
+    gchar buf[BUFSIZE];
+    glong len, curpos;
+    gsize bytes_read, bytes_written;
+    GError *error = NULL;
+    gchar *utf8str;
+
+    len = ImmGetCompositionString (himc, idx,
+                                   buf, G_N_ELEMENTS (buf));
+    curpos = ImmGetCompositionString(himc, GCS_CURSORPOS, NULL, 0);
+    utf8str = g_locale_to_utf8 (buf, len,
+                                &bytes_read, &bytes_written,
+                                &error);
+    if (error) {
+        g_warning ("%s", error->message);
+        g_error_free (error);
+    }
+
+    /* get cursor position by offset */
+    if (pos) {
+        /* FIXME! */
+        if (curpos < len && utf8str) {
+            gchar *tmpstr;
+            tmpstr = g_locale_to_utf8 (buf, curpos,
+                                       &bytes_read, &bytes_written,
+                                       &error);
+            if (error) {
+                g_warning ("%s", error->message);
+                g_error_free (error);
+            }
+            if (tmpstr) {
+                *pos = g_utf8_strlen (tmpstr, -1);
+            } else {
+                *pos = 0;
+            }
+            g_free (tmpstr);
+        } else if (curpos == len && utf8str) {
+            *pos = g_utf8_strlen(utf8str, -1);
+        } else {
+            g_warning ("ImmGetCompositionString: Invalid cursor position!");
+            *pos = 0;
+        }
+    }
+
+    return utf8str;
+}
+
+
 GdkFilterReturn 
 ime_message_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 {
@@ -232,21 +267,20 @@ ime_message_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 
     switch (msg->message) {
     case WM_IME_COMPOSITION:
-        if (msg->lParam & GCS_COMPSTR) {
-            g_signal_emit_by_name (context, "preedit_changed");
-        }
-
         if (msg->lParam & GCS_RESULTSTR) {
             gchar *utf8str;
 
-            GET_UTF8_PREEDIT_STRING (himc, GCS_RESULTSTR, utf8str);
+            context_ime->preediting = FALSE;
+            g_signal_emit_by_name (context, "preedit_changed");
+
+            utf8str = get_preedit_string (himc, GCS_RESULTSTR, NULL);
             if (utf8str) {
-#ifndef IMPLEMENT_GET_PREEDIT_STRING
                 g_signal_emit_by_name (G_OBJECT (context_ime),
                                        "commit", utf8str);
-#endif /* IMPLEMENT_GET_PREEDIT_STRING */
                 g_free (utf8str);
             }
+        } else if (context_ime->preediting && (msg->lParam & GCS_COMPSTR)) {
+            g_signal_emit_by_name (context, "preedit_changed");
         }
         break;
 
@@ -258,6 +292,7 @@ ime_message_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 
     case WM_IME_ENDCOMPOSITION:
         context_ime->preediting = FALSE;
+        g_signal_emit_by_name (context, "preedit_changed");
         g_signal_emit_by_name (context, "preedit_end");
         break;
 
@@ -364,11 +399,13 @@ gtk_im_context_ime_reset (GtkIMContext *context)
     if (context_ime->preediting && ImmGetOpenStatus (himc))
         ImmNotifyIME (himc, NI_COMPOSITIONSTR, CPS_CANCEL, 0);
 
+    context_ime->preediting = FALSE;
+    g_signal_emit_by_name (context, "preedit_changed");
+
     ImmReleaseContext (hwnd, himc);
 }
 
 
-#ifdef IMPLEMENT_GET_PREEDIT_STRING
 static void
 gtk_im_context_ime_get_preedit_string (GtkIMContext *context,
                                        gchar **str,
@@ -376,47 +413,51 @@ gtk_im_context_ime_get_preedit_string (GtkIMContext *context,
                                        gint *cursor_pos)
 {
     GtkIMContextIME *context_ime;
-    gchar *utf8str;
-    HWND hwnd;
-    HIMC himc;
-    gint utf8len;
-
-    g_return_if_fail (GTK_IS_IM_CONTEXT_IME (context));
 
     context_ime = GTK_IM_CONTEXT_IME (context);
 
-    hwnd = GDK_WINDOW_HWND (context_ime->client_window);
-    himc = ImmGetContext (hwnd);
+    if (context_ime->preedit_mode == GTK_IM_CONTEXT_IME_PREEDIT_GTK) {
+        gchar *utf8str;
+        HWND hwnd;
+        HIMC himc;
+        gint utf8len;
+        gint pos;
 
-    if (himc && context_ime->preediting) {
-        GET_UTF8_PREEDIT_STRING(himc, GCS_COMPSTR, utf8str);
+        hwnd = GDK_WINDOW_HWND (context_ime->client_window);
+        himc = ImmGetContext (hwnd);
+
+        if (himc && context_ime->preediting) {
+            utf8str = get_preedit_string (himc, GCS_COMPSTR, &pos);
+        } else {
+            utf8str = g_strdup("");
+            pos = 0;
+        }
+
+        utf8len = g_utf8_strlen (utf8str, -1);
+
+        if (str)
+            *str = utf8str;
+        else
+            g_free (utf8str);
+
+        if (attrs) {
+            /* When I add a underline attribute, Gtk+ often cause clash.
+               Is it Gtk's bug? Or my mistake?
+               foreground and background attributess are seems work correctly. */
+            *attrs = pango_attr_list_new ();
+        }
+
+        if (cursor_pos)
+            *cursor_pos = pos;
     } else {
-        utf8str = g_strdup("");
+        if (str)
+            *str = g_strdup ("");
+        if (attrs)
+            *attrs = pango_attr_list_new ();
+        if (cursor_pos)
+            *cursor_pos = 0;
     }
-
-    utf8len = g_utf8_strlen (utf8str, -1);
-
-    if (str)
-        *str = utf8str;
-    else
-        g_free (utf8str);
-
-#   warning FIXME
-#   if 1 /* FIXME!! */
-    if (attrs) {
-        PangoAttribute *attr;
-        *attrs = pango_attr_list_new ();
-        attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
-        attr->start_index = 0;
-        attr->end_index = utf8len > 0 ? utf8len - 1 : 0;
-        pango_attr_list_change (*attrs, attr);
-    }
-
-    if (cursor_pos)
-        *cursor_pos = 0;
-#   endif /* END FIXME!! */
 }
-#endif /* IMPLEMENT_GET_PREEDIT_STRING */
 
 
 static void
@@ -559,15 +600,20 @@ gtk_im_context_ime_set_cursor_location (GtkIMContext *context,
     if (!himc) return;
 
     if (ImmGetOpenStatus (himc)) {
-        POINT pt;
         COMPOSITIONFORM cf;
         gint wx = 0, wy = 0;
 
-        get_window_position (context_ime->client_window, &wx, &wy);
-        pt.x = wx + area->x;
-        pt.y = wy + area->y;
-        cf.dwStyle = CFS_POINT;
-        cf.ptCurrentPos = pt;
+        if (context_ime->preedit_mode == GTK_IM_CONTEXT_IME_PREEDIT_GTK) {
+            cf.dwStyle = CFS_RECT;
+            cf.rcArea.left  = cf.rcArea.top    = 0;
+            cf.rcArea.right = cf.rcArea.bottom = 0;
+        } else {
+            get_window_position (context_ime->client_window, &wx, &wy);
+            cf.dwStyle = CFS_POINT;
+            cf.ptCurrentPos.x = wx + area->x;
+            cf.ptCurrentPos.y = wy + area->y;
+        }
+
         ImmSetCompositionWindow (himc, &cf);
     }
 
