@@ -19,7 +19,7 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  *
- * $Id: gtkimcontextime.h,v 1.4 2003/05/07 10:08:31 makeinu Exp $
+ * $Id: gtkimcontextime.h,v 1.5 2003/05/09 14:05:06 makeinu Exp $
  */
 
 #include <gtk/gtkimcontext.h>
@@ -37,11 +37,17 @@ typedef struct _GtkIMContextIME        GtkIMContextIME;
 typedef struct _GtkIMContextIMEPrivate GtkIMContextIMEPrivate;
 typedef struct _GtkIMContextIMEClass   GtkIMContextIMEClass;
 
+typedef enum {
+    GTK_IM_CONTEXT_IME_PREEDIT_IME,
+    GTK_IM_CONTEXT_IME_PREEDIT_GTK,
+} GtkIMContextIMEPreeditMode;
+
 struct _GtkIMContextIME
 {
     GtkIMContext  object;
 
     GdkWindow    *client_window;
+    gint          preedit_mode;
     gboolean      preediting;
     gboolean      opened;
 
